(declaim (special *LTKTOOLS-DIR*))

(defun reader-hook (c e)
"Simple *DEBUGGER-HOOK* function. Used by LTKConsole."
  (declare (ignore c e))
  (invoke-restart (find-restart 'readtest)))

(defun ltkconsole-debug-hook (c e)
"Simple *DEBUGGER-HOOK* function. Used by LTKConsole."
  (declare (ignore e))
  (format t "DEBUG:~%  ~A~%" c)
  (invoke-restart (find-restart 'ltkeval)))

(defun display-history (twin hist-list disp-len)
"Display the history list to the LTKConsole widgets.
Used by LTKConsole."
  (ltk:clear-text twin)
  (let ((i 0)
	(len (if (> (length hist-list) disp-len)
		 disp-len
		 (length hist-list))))
    (dolist (n (subseq hist-list 0 len))
      (if (not (equal (third n) ""))
	  (progn
	    (with-open-file (fp "/tmp/ltk-foo.txt"
				:direction :output
				:if-exists :overwrite
				:if-does-not-exist :create)
	      (format fp "~A~%" (third n))
	      (force-output))
	    (ltk::load-and-append-text twin "/tmp/ltk-foo.txt")
	    (delete-file "/tmp/ltk-foo.txt")))
      (ltk:append-text twin
		       (concatenate 'string
				    (princ-to-string i)
				    "> "
				    (second n)))
      (incf i))))

(defun display-history-entry (twin foo)
"Display the history list to the LTKConsole widgets.
Used by LTKConsole."
  (ltk:clear-text twin)
  (if (not (equal (third foo) ""))
      (progn
	(with-open-file (fp "/tmp/ltk-foo.txt"
			    :direction :output
			    :if-exists :overwrite
			    :if-does-not-exist :create)
	  (format fp "~A~%" (third foo))
	  (force-output))
	(ltk::load-and-append-text twin "/tmp/ltk-foo.txt")
	(delete-file "/tmp/ltk-foo.txt")))
  (ltk:append-text twin (format nil "~A~%" (second foo))))

(defun ltkconsole ()
"LTKConsole - REPL Console using an LTK GUI
Provides a user interface to the Lisp REPL, with command line editing,
history, parenthesis auto-insert, and auto-completion."
  (declaim (special *LTKTOOLS-DIR* *HISTORY-LIST*))
  ;; We start the LTK interface with no debugging because it interferes
  ;; with our own. Maybe there is a way around this problem? 
  (let ((*HISTORY-LIST* nil)
	(history-idx 0))
    (ltk:with-ltk (:debug 0)
      (let* ((frame (make-instance 'ltk:frame))
	     (button-frame (make-instance 'ltk:frame :master frame))
	     (text-frame (make-instance 'ltk:frame :master frame))
	     (result (make-instance 'ltk:scrolled-text :master text-frame))
	     (bottom-frame (make-instance 'ltk:frame :master text-frame))
	     (eval-label (make-instance 'ltk:label :master bottom-frame :text "Eval: "))
	     (input (make-instance 'ltk:text :master bottom-frame))
	     
	     ;; The top row of buttons is defined in a user config file
	     (button-list (with-open-file (fp (concatenate 'string *LTKTOOLS-DIR* "/ltkconsole-buttons.lisp"))
			    (read fp)))
	     ;; Map the user config onto make-instance to make the widgets
	     (buttons (map 'list #'(lambda (x)
				     (make-instance 'ltk:button :master button-frame :text (car x)
						    :command (eval (cdr x)))) button-list)))
	
	(ltk:pack frame :fill :both :expand t)
	(ltk:pack button-frame)
	;; Pack our list of buttons
	(dolist (b buttons)
	  (ltk:pack b :side :right))
	(ltk:pack text-frame :fill :both :expand t)
	(ltk:pack result :fill :both :expand t)
	(ltk:pack bottom-frame :side :left :fill :both :expand t)
	(ltk:pack eval-label :side :left)

	(let ((opts (get-config-option :console-button)))
	  (dolist (b buttons)
	    (apply #'ltk:configure b opts)))

	(apply #'ltk:configure frame (get-config-option :console-frame))
	(apply #'ltk:configure text-frame (get-config-option :console-frame))
	(apply #'ltk:configure button-frame (get-config-option :console-frame))
	(apply #'ltk:configure bottom-frame (get-config-option :console-frame))

	(apply #'ltk:configure input (get-config-option :console-input))
	(apply #'ltk:configure (ltk:textbox result) (get-config-option :console-result))
	(apply #'ltk:configure eval-label (get-config-option :console-label))

	(ltk:pack input :fill :both :side :left :expand t)

	;; When the user presses Return/Enter we get ready to evaluate the 
	;; line that was entered. This is done by sandboxing the eval with a restart
	;; and a custom debugger hook. We also save the line and its result to the history.
	(ltk:bind input (get-config-option :console-key-eval)
		  (lambda (evt)
		    (declare (ignore evt))
		    (declaim (special *HISTORY-LIST*))
		    (let* ((fp (make-string-output-stream))
			   (*standard-output* fp)
			   (ret-eval "aborted"))
		      ;; Set up hook and restart
		      (let* ((*debugger-hook* #'ltkconsole-debug-hook))
			(with-simple-restart (ltkeval "LtkConsole Eval")
			  (setf ret-eval
				(let ((ret-list (multiple-value-list 
						 ;; Eval the line
						 (eval (read-from-string (ltk:text input)))))
				      (s ""))
				  (dolist (n ret-list s)
				    (setf s (concatenate 'string s 
							 (prin1-to-string n)
							 (make-string 1 :initial-element #\Newline))))))))
		      
		      (push (list (remove #\Newline (ltk:text input))
				  ret-eval
				  (get-output-stream-string fp))
			    *HISTORY-LIST*)
		      
		      (display-history (ltk:textbox result)
				       *HISTORY-LIST*
				       (get-config-option :display-length))
		      
		      (setf (ltk:text input) "")
		      (setf *HISTORY-LIST* (no-longer-than 
					    (get-config-option :history-length)
					    *HISTORY-LIST*))
		      (setf history-idx -1))))
	;; Revert to the previous history entry
	(ltk:bind input (get-config-option :console-key-previous)
		  (lambda (evt)
		    (declare (ignore evt))
		    (if (and (not (null *history-list*))
			     (< history-idx (1- (length *history-list*))))
			(let ((foo (elt *history-list* (1+ history-idx))))
			  (incf history-idx)
			  (setf (ltk:text input) (first foo))
			  (display-history-entry (ltk:textbox result)
						 foo)))))
	;; Proceed to the next history entry
	(ltk:bind input (get-config-option :console-key-next)
		  (lambda (evt)
		    (declare (ignore evt))
		    (cond 
		      ((> history-idx 0)
		       (decf history-idx)
		       (let ((foo (elt *history-list* history-idx)))						 
			 (setf (ltk:text input) (first foo))
			 (display-history-entry (ltk:textbox result)
						foo)))
		      ((= history-idx 0)
		       (display-history (ltk:textbox result)
					*HISTORY-LIST*
					(get-config-option :display-length))
		       (setf (ltk:text input) ""))
		      (t (setf history-idx -1)
			 (setf (ltk:text input) "")))))
	;; Auto-complete on the word under the text cursor
	(ltk:bind input (get-config-option :console-key-completer)
		  (lambda (evt)
		    (declare (ignore evt))
		    (let* ((idx (second (ltk::cursor-index input)))
			   (comp (auto-completer (ltk:text input) idx)))
		      (cond
			((listp comp)
			 (let ((*print-pretty* nil))
			   (setf (ltk:text (ltk:textbox result))
				 (princ-to-string comp))))
			((stringp comp)
			 (setf (ltk:text input) comp)))
		      (ltk:force-focus input))))
	;; Auto-insert a right parenthesis when given a left parenthesis
	(ltk:bind input "<KeyPress-parenleft>" (lambda (evt)
						 (declare (ignore evt))
						 (if (get-config-option :auto-paren)
						     (let* ((w (ltk:widget-path input))
							    (i (ltk::cursor-index input))
							    (r (first i))
							    (c (second i)))
						       (ltk:format-wish "~a insert ~a.~a {)}" w r c)
						       (ltk:format-wish "~a mark set insert {~a.~a}" w r c)))))
	;; Copy input line to the clipboard
	(ltk:bind input (get-config-option :console-key-copy)
		  (lambda (evt)
		    (declare (ignore evt))
		    (put-clip (read-from-string (ltk:text input)))))
	;; Paste the clipboard to the input line
	(ltk:bind input (get-config-option :console-key-paste)
		  (lambda (evt)
		    (declare (ignore evt))
		    (display-to-textbox input (get-clip))))
	;; Copy the result to the clipboard
	(ltk:bind (ltk:textbox result) (get-config-option :console-key-copy)
		  (lambda (evt)
		    (declare (ignore evt))
		    (put-clip (ltk:text (ltk:textbox result)))))
	;; Paste the clipboard to the result
	(ltk:bind (ltk:textbox result) (get-config-option :console-key-paste)
		  (lambda (evt)
		    (declare (ignore evt))
		    (display-to-textbox (ltk:textbox result) (get-clip))))))))


