
(defun shell-launch (program args)
"(SHELL-LAUNCH program args)
System independant wrapper to launch an OS program."
#+cmu (ext:run-program program args)
#+sbcl (sb-ext:run-program program args)
#+clisp (ext:run-program program :arguments args)
)

(defun get-list-of-threads ()
"(GET-LIST-OF-THREADS)
System independant wrapper to get a list of all running threads."
#+cmu mp::*all-processes*
#+sbcl (sb-thread:list-all-threads)
#+clisp nil
)

(defun get-name-of-thread (thr)
"(GET-NAME-OF-THREAD thr)
System independant wrapper to return the string name of a thread."
#+sbcl (sb-thread:thread-name thr)
#+cmu (mp:process-name thr)
#+clisp nil
)

(defun kill-thread (thr)
"(KILL-THREAD thr)
System independant wrapper to stop a running thread."
#+sbcl (sb-thread:destroy-thread thr)
#+cmu (mp:destroy-process thr)
#+clisp nil
)

(defun pause-thread (thr)
"(PAUSE-THREAD thr)
System independant wrapper to pause a running thread."
#+cmu (mp:disable-process thr)
#+sbcl nil
#+clisp nil
)

(defun enable-thread (thr)
"(ENABLE-THREAD thr)
System independant wrapper to enable a paused thread."
#+cmu (mp:enable-process thr)
#+sbcl nil
#+clisp nil
)

(defun get-slot-names (foo)
"(GET-SLOT-NAMES foo-class)
System independant wrapper to return all slot names of a class."
#+cmu  (pcl::get-slots foo)
#+sbcl (sb-pcl::get-slots foo)
#+clisp (clos::slot-names foo)
)

(defun display-to-textbox (textbox object)
"(display-to-textbox text-box object)
Displays the given object to the LTK textbox by writing it to a temporary
file, and loading it to the textbox with ltk:load-text."
  (with-open-file (fp "/tmp/foo.txt"
		      :direction :output
		      :if-exists :overwrite
		      :if-does-not-exist :create)
    (format fp "~A~%" object)
    (force-output fp))
  (ltk:load-text textbox "/tmp/foo.txt")
  (delete-file "/tmp/foo.txt"))

(defun list-dotted-length (L)
"Return the length of dotted list, ignoring the tail"
#+cmu  (lisp::dotted-list-length L)
#+sbcl (do ((R L (cdr R))
		(i 0 (1+ i)))
	       ((or (null R)
		    (atom R)) i))
#+clisp (ext:list-length-dotted L)
)
	
(defun fork-thread (fname sname)
"(fork-thread function-handle string-name)
System independant wrapper to launch a seperate thread."
#+cmu  (mp:make-process fname :name sname)
#+sbcl (sb-thread:make-thread fname :name sname)
#+clisp (funcall fname)
)

(defun classp (obj)
"(classp object)
Return T if the object is an instance of a class, NIL otherwise."
  (let ((foo (with-output-to-string (*standard-output*)
	       (describe obj))))
    (if (and
	 (search "class" foo)
	 (search "instance of" foo))
      t
      nil)))

(defun structp (obj)
"(structp object)
Return T if the object is a structure, NIL otherwise."
  (let ((foo (with-output-to-string (fp)
	       (describe-object obj fp))))
    (if (search "STRUCTURE-CLASS" foo)
	t
	nil)))

(defun get-struct-slots (S)
"(GET-STRUCT-SLOTS S)
Return a list of slot names for the given structure."
  (if (string-equal (subseq (princ-to-string S) 0 2) "#S")
      (let* ((list-form (read-from-string (subseq (princ-to-string S) 2)))
	     (new-list (make-list 1)))
	(dolist (L list-form (multiple-value-bind (x)
				 (read-from-string (remove #\: (princ-to-string (cdr new-list))))
			       x))
	  (if (keywordp L)
	      (nconc new-list (list L)))))
      nil))
    
(defun sref (obj &rest args)
"(sref object ..addy.. new-val)
References into object as with REF, and sets the element at the given address
to the new value (given as the last argument to the function)"
  (let ((addy (subseq args 0 (1- (length args))))
	(new-val (car (last args))))
  (cond
    ((and (arrayp obj)
	  (not (vectorp obj))
	  (eq (array-rank obj) (length addy)))
     (setf (apply #'aref obj addy) new-val))
    ((> (length addy) 1) 
     (let* ((new-addy (cond
		       ((or (listp obj)
 			    (vectorp obj)
 			    (hash-table-p obj)
	 		    (classp obj)
		 	    (structp obj))
			(list (car addy)))
 		       ((arrayp obj)
	 		(subseq addy 0 (array-rank obj)))))
	    (rest-addy (cond
			 ((or (listp obj)
			      (vectorp obj)
			      (hash-table-p obj)
			      (structp obj)
			      (classp obj))
			  (cdr addy))
			 ((arrayp obj)
			  (subseq addy (array-rank obj))))))
       (apply #'sref (apply #'ref obj new-addy) (append rest-addy (list new-val)))))
    ((= (length addy) 1)
     (cond
       ((and (listp obj)
	     (listp (cdr obj))
	     (> (list-dotted-length obj) (car addy)))
 	(setf (elt obj (car addy)) new-val))
       ((vectorp obj)
 	(setf (elt obj (car addy)) new-val))
       ((consp obj)
 	(if (= (car addy) 0)
	    (setf (car obj) new-val)
	    (setf (cdr (last obj)) new-val)))
       ((structp obj)
	(cond
	  ((and (symbolp (car addy))
		(search (list (car addy)) (get-struct-slots obj)))
	   (setf (slot-value obj (car addy)) new-val))
	  ((integerp obj)
	   (setf (slot-value obj (elt (get-struct-slots obj) (car addy)))
		 new-val))))
       ((classp obj)
	(cond
	  ((and (symbolp (car addy))
		(slot-boundp obj (car addy)))
	   (setf (slot-value obj (car addy)) new-val))
	  ((integerp (car addy))
           (setf (aref (get-slot-names obj) (car addy)) new-val)
	   )))
       ((hash-table-p obj)
	(if (numberp (car addy))
	      (let* ((keys (let ((tmp (make-list 1)))
			     (maphash #'(lambda (key val)
					  (declare (ignore val))
					  (nconc tmp (list key)))
				      obj)
			     (cdr tmp)))
		     (key (elt keys (car addy))))
		(setf (gethash key obj) new-val))
	      (setf (gethash (car addy) obj) new-val))))))))

(defun insref (obj &rest addy)
"(insref object ..addy..)
Inserts NIL into lists / arrays at the given address
Uses #\Space for strings
The address is dis-regarded for hash tables, but the new element is added."
  (cond
    ((null obj) nil)
    ((null addy) obj)
    ((null (car addy)) obj)
    ((listp obj)
     (let ((parent (apply #'ref obj (subseq addy 0 (1- (length addy)))))
 	   (last-addy (car (last addy))))
        (cond
	  ((listp parent)
	   (cond
	     ((eq last-addy 0)
	      (let ((foo-car (car parent))
		    (foo-cdr (cdr parent)))
		(setf (car parent) nil)
		(setf (cdr parent) (cons foo-car foo-cdr))))
	     ((eq last-addy (1- (length parent)))
	      (nconc parent (list nil)))
	     (t
	      (if (< last-addy (length parent))
		  (let ((foo parent)
			(tmp nil))
		    (dotimes (n last-addy)
		      (setf foo (cdr foo)))
		    (setf tmp (list (car foo)))
		    (setf (car foo) nil)
		    (setf (cdr foo) (cons (car tmp) (cdr foo))))))))
	  ((stringp parent)
	   (apply #'sref obj (append (subseq addy 0 (1- (length addy)))
				     (list (cond
					     ((eq last-addy 0)
					      (concatenate 'string " " parent))
					     ((eq last-addy (1- (length parent)))
					      (concatenate 'string parent " "))
					     (t
					      (concatenate 'string
							   (subseq parent 0 last-addy)
							   " "
							   (subseq parent last-addy))))))))
	  ((vectorp parent)
	   (apply #'sref obj (append (subseq addy 0 (1- (length addy)))
				     (list (cond
					     ((eq last-addy 0)
					      (concatenate 'vector #(nil) parent))
					     ((eq last-addy (1- (length parent)))
					      (concatenate 'vector parent #(nil)))
					     (t
					      (concatenate 'vector
							   (subseq parent 0 last-addy)
							   #(nil)
							   (subseq parent last-addy))))))))
	  ((hash-table-p parent)
	   (setf (gethash last-addy parent) nil)))))
    (t nil)))

(defun delref (obj &rest addy)
  "(delref object ..addy..)
  Delete the element at ADDY from the object, as with REF"
  (cond
    ((null obj) nil)
    ((null addy) obj)
    ((null (car addy)) obj)
    ((listp obj)
     (let ((parent (apply #'ref obj (subseq addy 0 (1- (length addy)))))
	   (last-addy (car (last addy))))
       (cond
	 ((listp parent)
	  (if (eq last-addy 0)
	      (progn
		(setf (first parent) (second parent))
		(delref parent 1))
	      (delete (ref parent last-addy)
		      parent
		      :start last-addy
		      :end (1+ last-addy))))
	 ((stringp parent)
	  (apply #'sref obj (append (subseq addy 0 (1- (length addy)))
				    (list (cond
					    ((eq last-addy 0)
					     (subseq parent 1))
					    ((eq last-addy (1- (length parent)))
					     (subseq parent 0 last-addy))
					    (t
					     (concatenate 'string
							  (subseq parent 0 last-addy)
							  (subseq parent (1+ last-addy)))))))))
	 ((vectorp parent)
	  (apply #'sref obj (append (subseq addy 0 (1- (length addy)))
				    (list (cond
					    ((eq last-addy 0)
					     (subseq parent 1))
					    ((eq last-addy (1- (length parent)))
					     (subseq parent 0 last-addy))
					    (t
					     (concatenate 'vector
							  (subseq parent 0 last-addy)
							  (subseq parent (1+ last-addy)))))))))
	 ((hash-table-p parent)
	  (if (and (integerp last-addy)
		   (>= last-addy 0))
	      (let* ((keys (let ((tmp (make-list 1)))
			     (maphash #'(lambda (key val)
					  (declare (ignore val))
					  (nconc tmp (list key)))
				      parent)
			     (cdr tmp)))
		     (key (elt keys last-addy)))
		(remhash key parent))
	      (remhash last-addy parent))))))
    (t nil)))

;; REF -- references a Lisp object with a uniform address
(defun ref (obj &rest addy)
"(ref obj ..addy..)
References a Lisp object with a uniform address."
  (cond
    ((null obj) nil)
    ((null addy) obj)
    ((null (car addy)) obj)
    ((vectorp obj)
     (apply #'ref (elt obj (car addy)) (cdr addy)))
    ((and (listp obj)
	  (listp (cdr obj))
	  (> (list-dotted-length obj) (car addy)))
     (apply #'ref (elt obj (car addy)) (cdr addy)))
    ((consp obj)
     (apply #'ref (if (= (car addy) 0)
	 	      (car obj)
 		      (cdr (last obj)))
	     (cdr addy)))
     ((arrayp obj)
      (if (>= (length addy) (array-rank obj))
 	 (apply #'ref (apply #'aref obj (subseq addy 0 (array-rank obj)))
 		(subseq addy (array-rank obj)))
 	 (apply #'ref (row-major-aref obj (car addy)) (cdr addy))))
     ((packagep obj)
      (cond
        ((symbolp (car addy))
 	(car addy))
        ((integerp (car addy))
 	(let ((foo (apropos-list "" (package-name obj))))
	    (elt foo (car addy))))))
     ((structp obj)
      (cond
        ((and (symbolp (car addy))
 	     (search (list (car addy)) (get-struct-slots obj)))
	 (apply #'ref 
	         (slot-value obj (car addy))
	        (cdr addy)))
        ((integerp (car addy))
 	(let* ((slots (get-struct-slots obj))
 	       (key (elt slots (car addy))))
 	  (apply #'ref (slot-value obj key)
 		 (cdr addy))))))
     ((classp obj) (cond
 		    ((and (symbolp (car addy))
 			  (slot-boundp obj (car addy)))
 		     (apply #'ref 
	 		    (slot-value obj (car addy))
	  		    (cdr addy)))
	 	    ((integerp (car addy))
	 	     (apply #'ref
	 		    (slot-value obj
		  			(aref (get-slot-names obj) (car addy)))
		 	    (cdr addy)))))
     ((hash-table-p obj) (cond 
 			  ((numberp (car addy))
 			   (apply #'ref 
	 			  (ref (let ((tmp (make-list 1))) 
	  				 (maphash #'(lambda (key val) 
	 					      (declare (ignore key)) 
	 					      (nconc tmp (list val))) 
	 					  obj)
		 			 (cdr tmp))
		  		       (car addy))
		 		  (cdr addy)))
	 		  (t (apply #'ref 
	 			    (gethash (car addy) obj)
		 		    (cdr addy)))))
     (t obj)))

;; Configure setf 
(defsetf ref sref)

(defun reflen (obj &rest addy)
"(reflen OBJ ..addy..) Returns the length of the element in OBJ at addy"
   (let ((foo (apply #'ref obj addy)))
     (cond
       ((and (listp foo)
	     (listp (cdr foo)))
        (if (not (null (cdr (last foo))))
 	   (1+ (list-dotted-length foo))
 	   (length foo)))
       ((vectorp foo)
        (length foo))
       ((packagep foo)
        (length (apropos-list "" (package-name foo))))
       ((consp foo) 2)
       ((arrayp foo) (array-total-size foo))
       ((hash-table-p foo)
	(length (let ((tmp (make-list 1))) 
		  (maphash #'(lambda (key val) 
			       (declare (ignore key)) 
			       (nconc tmp (list val))) 
			   foo)
		  (cdr tmp))))
       ((structp foo)
	(length (get-struct-slots foo)))
       ((classp foo)
	(length
	 (get-slot-names foo))))))

(defun atomic (foo)
  "Decides if an element is atomic.
Used for deciding if the object can be descended into further."
  (not (or (listp foo)
	   (arrayp foo)
	   (consp foo)
	   (structp foo)
	   (classp foo)
	   (hash-table-p foo)
	   (packagep foo))))

(defun expand-to-list (foo)
"Expand an object into a straight-list representation." 
  (cond
    ((listp foo) foo)
    ((vectorp foo) (map 'list #'(lambda (x) x) foo))
    ((arrayp foo) (let ((bar (make-list 1)))
		    (dotimes (n (array-total-size foo) (cdr bar))
		      (nconc bar (list (row-major-aref foo n))))))
    ((hash-table-p foo)
     (let ((bar (make-list 1)))
       (maphash #'(lambda (key val)
		    (declare (ignore key))
		    (nconc bar (list val))) 
		foo)
       (cdr bar)))
    ((structp foo)
     (map 'list #'(lambda (x) 
		    (declaim (special foo))
		    (slot-value foo x))
	  (get-struct-slots foo)))
    ((classp foo)
     (map 'list #'(lambda (x)
		    (declaim (special foo))
		    (slot-value foo x))
	  (get-slot-names foo)))
    ((packagep foo)
     (apropos-list "" (package-name foo)))))

(defun no-longer-than (len some-list)
"Returns a SEQUENCE that is not langer than LEN"
  (if (> (length some-list) len)
      (subseq some-list 0 len)
      some-list))

(defun tokenize-string (str token)
"Breaks a string into its fields given some character delimiter"
  (let ((tmp "")
	(ret-list '()))
    (dotimes (n (length str) (append ret-list (list tmp)))
      (if (eq (elt str n) token)
	  (progn
	    (setf ret-list (append ret-list (list tmp)))
	    (setf tmp ""))
	  (setf tmp (concatenate 'string tmp (princ-to-string (elt str n))))))))

(defun get-partial-word (str pos)
"Used by auto-completer. Finds the field number corresponding to a position in the text string."
  (let* ((words (tokenize-string str #\Space))
	 (max (cond
		((>= pos (length str)) (length str))
		((<= pos 0) 1)
		(t pos))))
    (do ((n 0 (+ n 1))
	 (len 0 (+ len (length (elt words n)) 1)))
	((>= len max) (1- n)))))

(defun auto-completer (str idx)
"Auto-completion 
Given a string 'str' and an index 'idx' auto-completer will find the space-delimited
word that is at the index position (via get-partial-word) and look this word up
via apropos-list to generate possible matches. Multiple possibilities are returned as
a symbol list, a single possibility is returned as a readable string."
  (let* ((all-words (tokenize-string str #\Space))
	 (n (get-partial-word str idx))
	 (word (remove #\Newline (remove #\) (remove #\( (copy-seq (elt all-words n))))))
	 (pack (let ((foo (tokenize-string word #\:)))
		 (if (> (length foo) 1)
		     (progn
			 (setf word (car (last foo)))
			 (read-from-string (first foo)))
		     nil)))
	 (a-matches (if pack
			(apropos-list word pack)
			(append (apropos-list word 'common-lisp)
				(apropos-list word 'common-lisp-user))))
	 (len (if pack
		  (length (format nil "~A:~A" pack word))
		  (length word)))
	 (matches  (let ((foo nil)
			 (*print-case* :downcase)
			 (*print-readably* (if pack
					       t
					       nil))
			 (*print-escape* t))
		     (dolist (m a-matches foo)
		       (if (and (not (search (list m) foo))
				(equal (subseq (princ-to-string m) 0 len)
				       (if pack
					   (format nil "~A:~A" pack word)
					   (princ-to-string word))))
			     (setf foo (append foo (list m))))))))
    (cond
      ((null matches) nil)
      ((> (length matches) 1) matches)
      (t (let ((*print-escape* t)
	       (*print-case* :downcase)
	       (ret ""))
	   (dotimes (i (length all-words) ret)
	     (if (eq n i)
		 (setf ret (concatenate 'string ret (paren-verify (remove #\newline (elt all-words i))
								  (if pack
								      (format nil "~A:~A" pack (car matches))
								      (car matches)))))
		 (setf ret (concatenate 'string ret (elt all-words i) " ")))))))))

(defun paren-verify (old new)
"Used by auto-completer. Takes two strings, old and new and insures that any parens on either side of old are
matched on new, and returns the string."
  (let ((begin (do ((i 0 (1+ i))
		    (f ""))
		   ((and (not (eq (elt old i) #\())
			 (not (eq (elt old i) #\)))) f)
		 (setf f (concatenate 'string f (princ-to-string (elt old i))))))
	(end (do ((i (1- (length old)) (1- i))
		  (f ""))
		 ((and (not (eq (elt old i) #\())
		       (not (eq (elt old i) #\)))) f)
	       (setf f (concatenate 'string f (princ-to-string (elt old i)))))))
    (format nil "~A~A~A" begin new (reverse end))))
